print("import: 'mypy'")
import mypy

print("import: 'mypy.dmypy'")
import mypy.dmypy

print("import: 'mypy.plugins'")
import mypy.plugins

print("import: 'mypy.server'")
import mypy.server

print("import: 'mypyc'")
import mypyc

